/* @(#)StageCtrkDialog.java
 *
 * vfstagecontroller java package: A simple way to drive a microscope serial
 * controlled stage.
 * Copyright 2003(C) Policlinico Universitario di Udine
 * Written by: Federico Viel
            vielfede@libero.it

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
package vfstagecontroller;
import javax.comm.*;
import java.lang.*;
import java.awt.*;
import java.io.*;
import java.util.*;


/**
 * <p>Title: Stage controller class </p>
 * <p>Description: By the use of this class the programmer shold be able to control a microscope stage (x,y,z). </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Policlinico universitario di Udine</p>
 * @author Federico Viel
 * @version 1.0
 */
public class StageController implements Runnable{
  /** To get x axis position*/
  static char X_COOR=1;
   /** To get y axis position*/
  static char Y_COOR=2;
  /** To get z axis position*/
  static char Z_COOR=3;

  /**Max number of Bookmarks*/
  public static int N_BOOKMARK=5;
  /**bookmark array*/
  private Bookmark bookmark[]=new Bookmark[N_BOOKMARK];
  /**Current bookmark index for insertion*/
  private static int bookmarkIndex;

  private boolean stage_stopped=false;

  private CommPortIdentifier portId;
  private SerialPort sPort;
  /**NAMES of SERIAL PORT(S) available on the computer */
  private String comPortName[]=new String[10];;
  private SerialParameters parametri;
  private SerialParameters parametriDiConnessione;
  private SerialConnection connessione;
  private String defaultComPortName;

  private TextArea commandArea = null;
  //private TextArea answerArea = new TextArea();
  static Enumeration portList;

  Dialog dlgToHide;

static {
bookmarkIndex=0;
}

  /**
   * Construct a stage controller setting default connection parameters and no command area:
   * "", 9600 baud,Flow control: NONE, 8 data bits, parity:NONE, 1 Stop bits
   * */
  public StageController() {
    initBookmarks();
    parametriDiConnessione=new SerialParameters("",9600,SerialPort.FLOWCONTROL_NONE,SerialPort.FLOWCONTROL_NONE,SerialPort.DATABITS_8,SerialPort.STOPBITS_1,SerialPort.PARITY_NONE);
    connessione=new SerialConnection(parametriDiConnessione);
    search4ComPortAvilable();
  }
  /**
   * Construct a stage controller setting default connection parameters:
   * "", 9600 baud,Flow control: NONE, 8 data bits, parity:NONE, 1 Stop bits
   * @param area echo commands  TextArea
   * */
  public StageController(TextArea area) {
    initBookmarks();
    parametriDiConnessione=new SerialParameters("",9600,SerialPort.FLOWCONTROL_NONE,SerialPort.FLOWCONTROL_NONE,SerialPort.DATABITS_8,SerialPort.STOPBITS_1,SerialPort.PARITY_NONE);
    connessione=new SerialConnection(parametriDiConnessione,area);
    search4ComPortAvilable();

  }
  /**
     * Construct a stage controller.
     * @param connPar connection parameters for serial connection with the stage controller
   * */
  public StageController(SerialParameters connPar) {
    initBookmarks();
    parametriDiConnessione=connPar;
    connessione=new SerialConnection(connPar);
    search4ComPortAvilable();

  }
  /**
       * Construct a stage controller.
       * @param connPar connection parameters for serial connection with the stage controller
       * @param area echo commands TextArea
     * */
    public StageController(SerialParameters connPar,TextArea area) {
      initBookmarks();
      parametriDiConnessione=connPar;
      connessione=new SerialConnection(connPar,area);
      search4ComPortAvilable();

  }

  /**
    * Search throught SERIAL & PARALLEL ports on board to the hardware saving JUST SERIAL port names on comPortName array.
    * @return default port for connection
    * @see #openConnection
    * @see #comPortName
   * */
  private String search4ComPortAvilable(){
    int i=0;
    defaultComPortName="";
    portList = CommPortIdentifier.getPortIdentifiers();
    while (portList.hasMoreElements()) {
           portId = (CommPortIdentifier) portList.nextElement();
           if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
                comPortName[i]=new String(portId.getName());
                if (i==0)
                  defaultComPortName=new String(comPortName[i]);
                i++;
           }
    }


    setDefaultPortName(defaultComPortName);
    return defaultComPortName;
  }
  /**
      * Gets default port for connection i.e. the port where connection will be attempted.
      * @return default port for connection
      * @see #openConnection
      *
   * */

  public String getDefaultPortName(){
    return defaultComPortName;
  }
  /**
    * Sets default port for connection i.e. the port where connection will be attempted.
    * @param dfpn SERIAL PORT NAME to be setted
    * @see #openConnection
    *
   * */
  public void setDefaultPortName(String dfpn){
    defaultComPortName=dfpn;
    parametriDiConnessione.setPortName(defaultComPortName);
  }

  /**
     * Estabilish connection with the controller and ping the controller to check its functionality
     * @return true if ping was ccompleted with happening, false otherwise
     * @throws SerialConnectionException
     * @see #moveStageTo
     * @see #closeConnection
     *
   * */
  public boolean openConnection() throws  SerialConnectionException{//Modificare il controllo delle eccezzioni...
    if (this.defaultComPortName.equals(""))
      throw new SerialConnectionException("No available com port selected");
    else{
      connessione.openConnection();
      return connessione.sendAck(2000);
    }

  }
  /**
   * Close the connection with the stage controller.
   * @see #openConnection
   * */
  public void closeConnection(){
    connessione.closeConnection();
  }

/**Get com port currently on board
 * @return com port list.
 * */
  public String[] getCommPortFounded(){
    return comPortName;
  }

  /**Get current serial connection parameters to estabilish serial connection with openConnection command
   * @return serial parameters
   * @see #openConnection
   * @see #setConnectionParameters
 * */
  public SerialParameters getConnectionParameters(){
    return parametriDiConnessione;
  }

  /**Set current serial connection parameters to estabilish serial connection with openConnection command
   * @param sp serial parameters
   * @see #openConnection
   * @see #getConnectionParameters
 * */
  public void setConnectionParameters(SerialParameters sp){
    parametriDiConnessione=sp;
  }
  /** Init bookmark array
   */
  public void initBookmarks(){
   for (int i=0;i<this.N_BOOKMARK;i++)
     bookmark[i]=new Bookmark("",-1,-1,-1);
  }
  /**Add a position bookmark on bookmark array. Bookmark array have a limeted allocation space.
   * This imply the insertion of bookmark on array space is circular
   * @param name Bookrmark name
   * @param x coordinate of the bookmark
   * @param y coordinate
   * @param z coordinate
   * @see getBookmarkName
   * @see getBookmarkCood
   * @return bookrmar  array index
   */
  public int addBookmark(String name, long x, long y, long z){
    bookmark[bookmarkIndex]=new Bookmark(name,x,y,z);
    bookmarkIndex=(bookmarkIndex+1)%N_BOOKMARK;
    return bookmarkIndex;
  }

  /** Retrive the bookmark name of position i
    * @param i bookmark position inside bookmark array
    * @return bookmark name
   */
  public String getBookmarkName(int i){
    String myarray[]=new String[this.N_BOOKMARK];
    return bookmark[i].name;
  }

  /**Return bookmark coordinate selected by kind param
   * @param name Bookmark name
   * @param kind x or y or z coordinate respecvely using X_COOR, Y_COOR, Z_COOR constant
   * @return -999999 if no right value can be returned
   */
  public long getBookmarkCoord(String name, char kind){
    int i=0;
    while (i< this.N_BOOKMARK)
      if (!name.equals(bookmark[i].name)){
         i++;
      }
      else{
        break;
      }

    if (i>=this.N_BOOKMARK)
      return -9999999;
    if (kind==this.X_COOR)
      return bookmark[i].xCoord;
    if (kind==this.Z_COOR)
      return bookmark[i].zCoord;
    if (kind==this.Y_COOR)
      return bookmark[i].yCoord;
    return -9999999;
  }

  /**
   * Get current position of requested axis:
   *  @param kind X_COOR return x position,  Y_COOR return y position, Z_COOR return z position
   *  @return x or y or z position
   * */
  synchronized public long getCurrentPos(char kind){
    String stringPos, X, Y, Z;
    int lastSpazio, firstSpazio;
    stringPos=connessione.sendCommand(StageCommands.GET_COORD);

    //System.out.println(stringPos);
    firstSpazio=0;
    lastSpazio=stringPos.indexOf(",");
    X=stringPos.substring(firstSpazio,lastSpazio);
    firstSpazio=lastSpazio+1;
    lastSpazio=stringPos.indexOf(",",firstSpazio);
    Y=stringPos.substring(firstSpazio,lastSpazio);
    firstSpazio=lastSpazio+1;
    lastSpazio=stringPos.indexOf("\n",firstSpazio);
    Z=stringPos.substring(firstSpazio,lastSpazio);
    if (kind==this.X_COOR)
      return Long.parseLong(X);
    if (kind==this.Z_COOR)
      return Long.parseLong(Z);
    if (kind==this.Y_COOR)
      return Long.parseLong(Y);
    return 0;
  }

  /**
   * Gets current X Step size.
   *  @return current x step size
 * */
  synchronized public long getXStepSize(){
    String stringPos, X;
    int lastSpazio, firstSpazio;
    String command=StageCommands.builtCommand(StageCommands.Y_STEP_SIZE);
    stringPos=connessione.sendCommand(command);
    firstSpazio=0;
    lastSpazio=stringPos.indexOf(",");
    X=stringPos.substring(firstSpazio,lastSpazio);
    return Long.parseLong(X);
  }


  /**
   * Gets current Y Step size.
   *  @return current y step size
 * */
  synchronized public long getYStepSize(){
    String stringPos, Y;
    int lastSpazio, firstSpazio;
    String command=StageCommands.builtCommand(StageCommands.X_STEP_SIZE);
    stringPos=connessione.sendCommand(command);
    firstSpazio=stringPos.indexOf(",")+1;
    lastSpazio=stringPos.indexOf("\n",firstSpazio);
    Y=stringPos.substring(firstSpazio,lastSpazio);
    return Long.parseLong(Y);
  }



  /**
   * Gets Z Step size in decimal micron.
   * @return current Z step size
   */
  synchronized public long getZStepSize(){
    String command,r1,rvalue;
    command=StageCommands.builtCommand(StageCommands.Z_STEP_SIZE);

    r1=connessione.sendCommand(command);
    rvalue=r1.substring(0,r1.indexOf("\n"));
    return Long.parseLong(rvalue);
  }

  /**
     * Sets X-Step size of stage in micron.
     * @param v step size in micron
   * */
  synchronized public void setXStepSize(long v){
    String command=StageCommands.builtCommand(StageCommands.X_STEP_SIZE,v,getYStepSize());
    connessione.sendCommand(command);
  }
  /**
     * Set Y-Step size of stage in micron.
     * @param v step size in micron
   * */
  synchronized public void setYStepSize(long v){
    String command=StageCommands.builtCommand(StageCommands.X_STEP_SIZE,getXStepSize(),v);
    connessione.sendCommand(command);
  }
  /**
   * Set X and Y Step size of stage in micron.
   * @param u X-Step size in micron
   * @param v Y-Step size in micron
   * @return stage response string for error control
   * */
 synchronized  public String setXYStepSize(long u, long v){
    String command=StageCommands.builtCommand(StageCommands.X_STEP_SIZE,u,v);
    return connessione.sendCommand(command);
  }

  /**
   * Set Z-Step size of stage in micron.
   * @param v focus step size in micron
   * @return stage response string for error control
   * */
  synchronized public String setZStepSize(long v){
  String command=StageCommands.builtCommand(StageCommands.Z_STEP_SIZE,v);
  return connessione.sendCommand(command);
  }

  /**
  * Set X, Y, Z Step size of stage in micron.
  * @param u X-Step size in micron
  * @param v Y-Step size in micron
  * @param w Z-Step size in micron
  * @return stage response string for error control
   * */
  synchronized public String setXYZStepSize(long u, long v, long w){
    return setXYStepSize(u,v)+setZStepSize(w);
  }



  /**
   * Get Backlash stage X-Y status.
   * @return X-Y Backlash status:
   * false: Backlash disabled
   * true:  Backlash enabled
   */
  synchronized public boolean getBacklashStageStat(){
  String command=StageCommands.builtCommand(StageCommands.STAGE_BACKLASH_H);
  String r1, rvalue ;
  r1=connessione.sendCommand(command);
  rvalue=r1.substring(0,r1.indexOf(","));
  //System.out.print(rvalue);
  if (Integer.parseInt(rvalue)==0)
    return false;
  else
    return true;
  }

  /**
   * Get Backlash focus status.
   * @return Z-Backlash status:
   * false: Z-Backlash disabled
   * true:  Z-Backlash enabled
   */
  synchronized public boolean getBacklashFocusStat(){
    String command=StageCommands.builtCommand(StageCommands.Z_BACKLASH_H);
    String r1, rvalue;
    r1=connessione.sendCommand(command);
    rvalue=r1.substring(0,r1.indexOf(","));
    if (Integer.parseInt(rvalue)==0)
      return false;
    else
      return true;
  }

  /**
   * Enable or disable X,Y stage axis backlash
   * @param b true=enable, false=disable
   * @return stage response string for error control
   */
  synchronized public String enableStageBackLash(boolean b ){
    String command;
    long s;
    if (b)
      s=1;
    else
      s=0;
    command=StageCommands.builtCommand(StageCommands.STAGE_BACKLASH_H,s,10000);
    return connessione.sendCommand(command);
  }
  /**
   * Enable or disable Z stage axis backlash
   * @param b true=enable, false=disable
   * @return stage response string for error control
   */
  synchronized public String enableZBackLash(boolean b ){
    String command;
    long s;
    if (b)
      s=1;
    else
      s=0;
    command=StageCommands.builtCommand(StageCommands.Z_BACKLASH_H,s,10000);
    return connessione.sendCommand(command);
  }
  /**
  * Move Stage to Absolute position x,y,z
  * @param x absolute X position in micron
  * @param y absolute Y position in micron
  * @param z absolute Z position in micron
  */
  synchronized public void moveStageTo(long x, long y,long z){
    moveStageTo(x,y);
    moveZAxisTo(z);
  }

  /**
  * Move Z to Absolute position z
  * @param z absolute Z position in micron
  */
  synchronized public void moveZAxisTo(long z){
   String command;
   command=new String(StageCommands.MOVE_Z_TO_ABS_POS+" "+String.valueOf(z));
   connessione.sendCommand(command);
  }
  /**
  * Move stage to absolute position x,y
  * @param x absolute X position in micron
  * @param y absolute Y position in micron
  * @return string stage response
  */
  synchronized public String moveStageTo(long x, long y){
    String command;
    command=new String(StageCommands.MOVE_TO_ABS_POS+" "+String.valueOf(x)+" "+String.valueOf(y));
    return connessione.sendCommand(command);
  }

  /**
  * Move stage in the selected direction by the size defined in the step dim:
  * @param command movement direction:
  *   StageCommands.MOVE_DOWN: move the stage a step down
  *   StageCommands.MOVE_UP:   move the stage a step up
  *   StageCommands.MOVE_RIGHT:move the stage a step right
  *   StageCommands.MOVE_LEFT: move the stage a step left
  *   @return string stage response
  */
  synchronized public String moveXYStep(String command){
    return connessione.sendCommand(command);
  }
  /**
   * Move Z axis by one step:
   * @param command movement direction:
   * StageCommands.MOVE_Z_UP    move z up by one step
   * StageCommands.MOVE_Z_DOWN  move z down by one step
     * @return string stage response
*/
  synchronized public String moveZStep(String command){
    return connessione.sendCommand(command);
  }

  /**
   * Set current stage position to absolute position x,y,z.
   * @param x new absolute X position (in micron) value for current X-Stage position
   * @param y new absolute Y position (in micron) value for current Y-Stage position
   * @param z new absolute Z position (in micron) value for current Z-Stage position
   *   @return string stage response
   */
  synchronized public String setAbsPos(int x, int y, int z){
  String command=StageCommands.builtCommand(StageCommands.SET_ABS_POS,x,y,z);
  return connessione.sendCommand(command);
  }
  /** Launch searchSetLimits method.
   * serchSetLimits() have to run in async mode to give to the user the possibility of process stopping.
   */
  public void run(){
    searchSetLimits();
  }
  /**
   * Automatic stage fisical limits search function: 1) go to upper left corner; 2) Check the position;
   * 3) go to bottom right position; 4) check the position; 5) Calculate the middle position and go there.
   * This procedure uses StageCommands.X_LIMIT and StageCommands.Y_LIMIT constants.
   */
  private void searchSetLimits(){
    long r,b,mx,my;
    stage_stopped=false;
    setMode(0);
    moveStageTo(-StageCommands.X_LIMIT,-StageCommands.Y_LIMIT);
    while (getStatus()!=0)
      if (stage_stopped)
        return;
    setMode(1);
    this.getCurrentPos(this.X_COOR);
    this.getCurrentPos(this.Y_COOR);
    setAbsPos(0,0,0);
    setMode(0);
    moveStageTo(StageCommands.X_LIMIT,StageCommands.Y_LIMIT);
    while (getStatus()!=0)
      if (stage_stopped)
        return;
    setMode(1);
    r=this.getCurrentPos(this.X_COOR);
    b=this.getCurrentPos(this.Y_COOR);
    mx=(long)(r/2);
    my=(long)(b/2);

    setMode(0);
    moveStageTo(mx,my);
    while (getStatus()!=0)
      if (stage_stopped)
        return;
    setMode(1);
    setAbsPos(0,0,0);
    dlgToHide.setVisible(false);
    return;
  }

  /**
   * Set stage speed.
   * @deprecated No enougth information on stage protocol manual
   * @param speed speed
   */
  synchronized public void setStageSpeed(int speed){
    if (speed<StageCommands.STAGE_MAX_SPEED){
      String command=StageCommands.builtCommand(StageCommands.STAGE_SPEED,(long) -speed, (long) speed);
      connessione.sendCommand(command);
    }
  }
  /**
   * Get stage speed range.
   * @return x,y  stage speed range
   */
  synchronized public int getXYSpeed(){
    String r1,rvalue;
    String command=StageCommands.builtCommand(StageCommands.XY_SPEED);
    r1=connessione.sendCommand(command);
    rvalue=r1.substring(0,r1.indexOf("\n"));
    return Integer.parseInt(rvalue);
  }
  /**
   * Set stage XY-axis speed range.
   * set X-Y Stage speed range
   * @param speed speed
   * @return stage response string for error control
   */
  synchronized public String setXYSpeed(int speed){
    String command=StageCommands.builtCommand(StageCommands.XY_SPEED,(long) speed);
    return connessione.sendCommand(command);
  }
  /**
   * Return Stage XY acceleration.
   * @return X-Y acceleration
 */
  synchronized  public int getXYAcc(){
    String r1,rvalue;
    String command=StageCommands.builtCommand(StageCommands.XY_ACC);
    r1=connessione.sendCommand(command);
    rvalue=r1.substring(0,r1.indexOf("\n"));
    return Integer.parseInt(rvalue);
  }
  /**
   * Set stage X,Y-axis acceleration.
   * @param acc X-Y acceleration
   * @return stage response string for error control
   */
  synchronized public String setXYAcc(int acc){
  String command=StageCommands.builtCommand(StageCommands.XY_ACC,(long) acc);
  return connessione.sendCommand(command);
  }

  /**
   * Get z axis speed range.
   * @return Z stage speed range
   */
  synchronized public int getZSpeed(){
    String r1,rvalue;
    String command=StageCommands.builtCommand(StageCommands.Z_SPEED);
    r1=connessione.sendCommand(command);
    rvalue=r1.substring(0,r1.indexOf("\n"));
    return Integer.parseInt(rvalue);
  }
  /**
   * Set stage Z-axis speed range.
   * set Z Stage speed range
   * @param speed new speed
   * @return stage response string for error control
   */
  synchronized public String setZSpeed(int speed){
    String command=StageCommands.builtCommand(StageCommands.Z_SPEED,(long) speed);
    return connessione.sendCommand(command);
  }
  /**
   * Return Z axis acceleration.
   * @return Z acceleration
 */
 synchronized  public int getZAcc(){
    String r1,rvalue;
    String command=StageCommands.builtCommand(StageCommands.Z_ACC);
    r1=connessione.sendCommand(command);
    rvalue=r1.substring(0,r1.indexOf("\n"));
    return Integer.parseInt(rvalue);
  }
  /**
   * Set stage Z-axis acceleration.
   * @param acc new Z acceleration
   * @return stage response string for error control
   */
  synchronized public String setZAcc(int acc){
  String command=StageCommands.builtCommand(StageCommands.Z_ACC,(long) acc);
  return connessione.sendCommand(command);
  }
  /** Stop the stage if is moving. To work this methos needs that stage is in "compatibility mode"
   * @return return command response if the operation had success null otherwise
   */
  synchronized public String stopStage(){
   String rvalue;
    if (getMode()==0){
      String command=StageCommands.builtCommand(StageCommands.STOP_STAGE);
      rvalue=connessione.sendCommand(command);
      stage_stopped=true;
      setMode(1);
      return rvalue;
    }
    return null;
  }
  /** Send a request to stage controller questioning about its mode: usually COMPATIBILITY or STANDARD.
   * May be this method could be VERY different if stage serial protocol is changed (i.e. when you would drive another stage)
   * @return mode (integer value)
   */
  synchronized public int getMode(){
    String r1,rvalue;
    String command=StageCommands.builtCommand(StageCommands.MODE);
    r1= connessione.sendCommand(command);
    rvalue=r1.substring(0,r1.indexOf("\n"));
    return Integer.parseInt(rvalue);
  }
  /**Send to the stage the request to swap in mode m
   * @param m mode
   * @return stage response: null if m is different from allowed values
   */
  public String setMode(int m){
    String command;
    if (m==0 |m==1){
      command=StageCommands.builtCommand(StageCommands.MODE,m);
      return connessione.sendCommand(command);
    }
    return null;
  }
  /**Get status of the stage. Only available in compatibility mode. See prior documentation for more informations
   * @return code value of stage state.
   */
 synchronized  public int getStatus(){
    String r1,rvalue;
    String command=StageCommands.builtCommand(StageCommands.GET_STATUS);
    r1= connessione.sendCommand(command);
    rvalue=r1.substring(0,r1.indexOf("\n"));
    return Integer.parseInt(rvalue);
  }
  /** Set the dialog to hide at the end of seachLimits methods
   * @param d the dialog
   * @see searchSetLimits
   */
  public void setDialogToHide(Dialog d){
    dlgToHide=d;
    return;
  }
}